############################################################
#
# Speech Recognition Sub Module for Julius
#
#  by speech recognition group, IPA agent project
#
#  Ver.0.0  (2001/06/25) (internal) initial version, "inq Options" added
#  Ver.0.1  (2001/06/??) (internal) initial interaction with julian
#  Ver.0.2  (2001/07/04) (internal) alpha release
#  Ver.0.3  (2001/07/09) initial public release
#  Ver.0.4  (2002/01/29) (internal) support for switching grammars
#  Ver.0.5  (2002/02/11) (internal) support for XML-format grammar ; added by A.Kai
#  Ver.0.6  (2003/01/30) update with julius-3.3p3-Galatea
#  Ver.0.9  (2005/07/07) run with julius-3.5-rc5 with gmm rejections and more.
#  Ver.0.9.2(2006/07/31) run with julius-3.5.2
#  Ver.1.0.0(2008/08/07) run with Julius-4.0.2  (julian -> julius)
#
# developed on perl version 5.005_03 built for i386-linux
#

use strict;
use IO::Socket;
use Fcntl;

my($version) = "1.0.0";

############################################################
#
# site defaults
#

# read in config variables
require "config.pl";
use vars qw ( %CONF );

# Julian executable prefix
my($julianbin_prefix) = $CONF{"JULIAN_BIN_PREFIX"};
# raw output log of julian 
my($logfile) = $CONF{"LOG_JULIAN"};
# (for debug) set to 1 to see julian raw output to stdout
my($debug) = $CONF{"DEBUG_LEVEL"};

# Sub module for grammar format converter (XML -> Julian)
my( $xml2julgrm ) = $CONF{"XML2JULGRM"};
my( $xmldtd) = $CONF{"XML_DTD"};

require $xml2julgrm;

############################################################
#
# global variables
#

# these values are shared with SRM_Main.pl
%SLOT = ();	# slot
%PROPERTY = ();	# slot property 

# default option values (switch options has value of ON/OFF)
use vars qw ( %binval );
use vars qw ( %optval );

# correspondence of "set" command to select Julian executable binaries
#my( %set2bin ) = ("Input.micapi" => "API:{oss|alsa|esd}",
#		  "Output.format" => "GRAPH:{nbest|graph}");
#my( $julian_suffix_template ) = "-API-GRAPH";
my( %set2bin ) = ();
my( $julian_suffix_template ) = "";

# correspondence of "set" commands to Julian option name (with arg type)
my( %set2opt ) = ("AcousticModel.hmm" => "-h:file",
		  "AcousticModel.imp" => "-hlist:file",
		  "GMM.model" => "-gmm:file",
		  "GMM.calcnum" => "-gmmnum:int",
		  "GMM.rejectstr" => "-gmmreject:str",
		  "Output.multigram" => "-multigramout:sw",
		  "Grammar" => "-dfa:file",
		  "Grammer" => "-dfa:file", # ver.0 spec. typo
		  "Dic" => "-v:file",
# 06/02/03 by ri: added charcode specification for grammar
		  "Grammar.charcode" => "-charconv_in:{euc-jp|euc|sjis|utf-8|default}",
		  "Param.beam1" => "-b:int",
		  "Param.beam2" => "-b2:int",
		  "Param.stack" => "-s:int",
		  "Param.overflow" => "-m:int",
		  "Param.lookuprange" => "-lookuprange:int",
# 06/02/03 by ri: add looktrellis command
		  "Param.looktrellis" => "-looktrellis:sw",
		  "Param.gauss" => "-tmix:int",
		  "Param.rejectshort" => "-rejectshort:intsw",
		  "Output.nbest" => "-n:int",
		  "Output.word" => "-outcode:sw_W",
		  "Output.LMword" => "-outcode:sw_L",
		  "Output.phone" => "-outcode:sw_P",
		  "Output.score" => "-outcode:sw_S",
		  "Output.cm" => "-outcode:sw_C",
		  "Output.pass1word" => "-outcode:sw_w",
		  "Output.pass1LMword" => "-outcode:sw_l",
		  "Output.pass1phone" => "-outcode:sw_p",
		  "Output.pass1score" => "-outcode:sw_s",
		  "Output.pass1prog" => "-progout:sw", 
		  "Output.walign" => "-walign:sw", 
		  "Output.palign" => "-palign:sw", 
		  "Output.pass1progtime" => "-proginterval:int",
# 06/02/03 by ri: kanjicode -> charconv
#		  "Output.kanjicode" => "-kanji:{euc|sjis|jis|noconv}",
		  "Output.kanjicode" => "-charconv_out:{euc-jp|euc|sjis|jis|utf-8|noconv}",
		  "Output.charcode" => "-charconv_out:{euc-jp|euc|sjis|jis|utf-8|noconv}",
		  "Output.lattice" => "-lattice:sw",
		  "Record.dir" => "-record:dir",
		  "Input.from" => "-input:{mfcfile|rawfile|mic|adinnet|netaudio|oss|alsa|esd}",
		  "Input.zmean" => "-zmeanframe:sw",
		  "Input.level" => "-lv:int");

# correspondence of "inq" command to Julian executable binary selecter
#my( %inq2bin ) = ("Input.micapi" => "API",
#		  "Output.format" => "GRAPH");
my( %inq2bin ) = ();

# correspondence of "inq" commands to Julian option name
my( %inq2opt ) = ("AcousticModel.hmmfile" => "-h",
		  "AcousticModel.impfile" => "-hlist",
		  "GMM.model" => "-gmm",
		  "GMM.calcnum" => "-gmmnum",
		  "GMM.rejectstr" => "-gmmreject",
		  "Output.multigram" => "-multigramout",
		  "Grammar.file" => "-dfa",
		  "Grammer.file" => "-dfa", # ver.0 spec. typo
		  "Grammar.type" => "-dummy", # added by A.Kai
		  "Dic.file" => "-v",
		  # Dic.type
# 06/02/03 by ri: added charcode specification for grammar
		  "Grammar.charcode" => "-charconv_in",
		  "Param.beam1" => "-b",
		  "Param.beam2" => "-b2",
		  "Param.stack" => "-s",
		  "Param.overflow" => "-m",
		  "Param.lookuprange" => "-lookuprange",
# 06/02/03 by ri: add looktrellis command
		  "Param.looktrellis" => "-looktrellis",
		  "Param.gauss" => "-tmix", # for PTM only
		  "Param.rejectshort" => "-rejectshort",
		  "Output.nbest" => "-n",
		  "Output.word" => "-outcode:sw_W",
		  "Output.LMword" => "-outcode:sw_L",
		  "Output.phone" => "-outcode:sw_P",
		  "Output.score" => "-outcode:sw_S",
		  "Output.cm" => "-outcode:sw_C",
		  "Output.pass1word" => "-outcode:sw_w",
		  "Output.pass1LMword" => "-outcode:sw_l",
		  "Output.pass1phone" => "-outcode:sw_p",
		  "Output.pass1score" => "-outcode:sw_s",
		  "Output.pass1prog" => "-progout", 
		  "Output.walign" => "-walign", 
		  "Output.palign" => "-palign", 
		  "Output.pass1progtime" => "-proginterval", # not implemented yet
# 06/02/03 by ri: kanjicode -> charconv
#		  "Output.kanjicode" => "-kanji",
		  "Output.kanjicode" => "-charconv_out",
		  "Output.charcode" => "-charconv_out",
		  "Output.lattice" => "-lattice",
		  "Record.dir" => "-record",
		  "Input.from" => "-input",
		  "Input.zmean" => "-zmeanframe",
		  "Input.level" => "-lv");

# response to common inquery (except "Run") and their values
my ( %commoninq ) = ("ProtocolVersion" => "Protocol Ver. 1",
		     "ModuleVersion" => "Speech Recognition Module (Julius) Ver. $version");

# global variables for grammar types; added by Kai.
my( %grammartype ) =  ( ".dfa" => "Julian-DFA",
			".xml" => "IPA-SRG-XML"
			);
# default grammar file prefix
my( $default_grammarprefix ) = $CONF{"TMP_GRAMMAR_PREFIX"};

#############################################################
#
# misc global values
#

my ( $module_status ) = "DEAD";	# module status
my ( $julian_socket );		# accepted socket to communicate with julian
my ( $jport ) = $CONF{"JPORT"};		# comminucation port number
my ( $jpid );			# process ID of julian
my ( %modified );
my ( $module_started);

#############################################################
#
# handle "set" command
# 
# return value: 0 for error, other for success
#
# when this function is called from SRM_Main.pl, the slot value
# "$SLOT{$slot}" is already set to $value.
# you can modify the slot value within this script.
#
sub SubCall_Set{
    my($slot, $value) = @_;
    my($optstr, $type, $grmhead);
    my($xml_encoding);

    if ($set2bin{$slot} ne "") {       # julian executable selection
	($optstr, $type) = split(/:/, $set2bin{$slot});
	if ($type =~ /^\{.*\}$/) {	# selection arg
	    if (grep(/^($value)$/, split(/[\{\}\|]/, $type)) == 0) {
		print_tell("invalid value $value\n");
		return 0;
	    }
	    $binval{$optstr} = $value;
	}
    } elsif ($set2opt{$slot} ne "") {	# julian option values
	if ($set2opt{$slot} eq "??") { # not implemented yet
	    print_tell("not implemented slot $slot\n");
	    return 0;
	}
	($optstr, $type) = split(/:/, $set2opt{$slot});
	if ($type eq "file") {	# file arg
#	    if (! -f $value) {	# not exist
#		print_tell("no such file $value\n");
#		return 0;
#	    }
	    $optval{$optstr} = $value;
	    if ($slot eq "AcousticModel.hmm") {
		$SLOT{"AcousticModel.hmmfile"} = $value;
	    }
	    if ($slot eq "AcousticModel.imp") {
		$SLOT{"AcousticModel.impfile"} = $value;
	    }
	    if ($slot eq "Grammar" || $slot eq "Grammer") {
		$value =~ /^\s*([^\s]*)\s*/;
		$grmhead = $1;
		if ($grmhead =~ /\.dfa$/){
		    $SLOT{"Grammar.type"} = $grammartype{".dfa"};
		}
		elsif ($grmhead =~ /\.xml$/ || $grmhead =~ /^<\?xml/i){
		    $SLOT{"Grammar.type"} = $grammartype{".xml"};
	            $SLOT{"Dic.file"} = "";
		}
		else {
		    print_tell("unsupported speech recognition grammar format: $grmhead\n");
		    return 0;
		}
		if ($grmhead =~ /^<\?xml/i){
		    $grmhead = $default_grammarprefix . ".xml";
		    if ( !open(XML, "> $grmhead") ){
			print_tell("cannot open internal XML-grammar file: $grmhead\n");
			return 0;
		    }
		    print XML $value;
		    close( XML );
		    # 02/02/22 by ri: copy DTD to temp
		    system("cp $xmldtd `dirname $default_grammarprefix`");
	            $SLOT{"Grammar.file"} = $grmhead;
		}
		else {
	            $SLOT{"Grammar.file"} = $value;
		}
		if ($SLOT{"Grammar.type"} eq $grammartype{".xml"}) {
# 06/02/03 by ri: get charset of XML grammar
		    if (!open(XML, "$grmhead")) {
			print_tell("cannot open internal XML-grammar file: $grmhead\n");
			return 0;
		    }
		    $xml_encoding="";
		    while(<XML>) {
			if (/\<\?xml.*encoding=\"(.*)\"/) {
			    $xml_encoding = $1;
			    last;
			}
		    }
		    close(XML);
		    $xml_encoding =~ tr/A-Z/a-z/; # to lower case
		    $optval{"-charconv_in"} = $xml_encoding;

		    $grmhead =~ s/(.*)\.xml/$1/;
		    print_tell("grammar is now converted to Julian format...\n");
		    if (&convertGrammar($grmhead) != 0){
			print_tell("grammar conversion failed.\n");
			return 0;
		    }
		    print_tell("grammar conversion complete.\n");
		    $optval{$optstr} = $grmhead . ".dfa";
		    print_debbug("optval $optstr =" . $optval{$optstr} . "\n");
		    ($optstr, $type) = split(/:/, $set2opt{"Dic"});
		    $optval{$optstr} = $grmhead . ".dict";
		    print_debbug("optval $optstr =" . $optval{$optstr} . "\n");
		    if (defined($julian_socket)) {
		        $modified{"grammar"} = 1;
		        $modified{"dict"} = 1;
		        &change_gram_julian;
		    }
		}
		elsif (defined($julian_socket)) {
		    $modified{"grammar"} = 1;
		    &change_gram_julian;
		}
	    }
	    if ($slot eq "Dic") {
		$SLOT{"Dic.file"} = $value;
		if (defined($julian_socket)) {
		    $modified{"dict"} = 1;
		    &change_gram_julian;
		}
	    }
	} elsif ($type eq "dir") { # directory or OFF
	    if ($value ne "OFF") {
		if (! (-d $value && -w $value)) {
		    print_tell("cannot write to directory \"$value\"\n");
		    return 0;
		}
	    }
	    $optval{$optstr} = $value;
	} elsif ($type eq "int") { # integer arg
	    $optval{$optstr} = $value;
	} elsif ($type eq "str") { # string arg
	    $optval{$optstr} = "$value";
	} elsif ($type eq "sw") { # for option with no arg: ON / OFF
	    if ($value eq "ON" || $value eq "on") {
		$optval{$optstr} = "ON";
	    } elsif ($value eq "OFF" || $value eq "off") {
		$optval{$optstr} = "OFF";
	    } else {
		print_tell("invalid value $value\n");
		return 0;
	    }
	} elsif ($type eq "intsw") { # integer arg(=ON) or OFF
	    $optval{$optstr} = $value;
	} elsif ($type =~ /^sw_(.*)$/) { # add/remove char (for "-outcode")
	    my ($val) = $1;
	    if ($value eq "ON" || $value eq "on") {
		unless ($optval{$optstr} =~ /$val/) {
		    $optval{$optstr} .= $val;
		}
	    } elsif ($value eq "OFF" || $value eq "off") {
		$optval{$optstr} =~ s/$val//g;
	    } else {
		print_tell("invalid value $value\n");
		return 0;
	    }
	} elsif ($type =~ /^\{.*\}$/) {	# selection arg
	    if (grep(/^($value)$/, split(/[\{\}\|]/, $type)) == 0) {
		print_tell("invalid value $value\n");
		return 0;
	    }
	    $optval{$optstr} = $value;
	}
    } elsif ($slot eq "Run") {
	if($value eq "INIT"){ # prepare for service 
	    # does not "LIVE" before initialize success (2002/02/11 ri)
	    #$module_status = "LIVE";
	} elsif ($value eq "EXIT"){ # stop for service
	    &stop_julian;	# terminate julian process
	    $module_status = "DEAD";
	} elsif ($value eq "START") {
	    $module_status = "LIVE"; # ensure the module is alive
	    &start_julian;	# spawn julian
	} elsif ($value eq "PAUSE") {
	    &pause_julian;
	} elsif ($value eq "TERMINATE") {
	    &terminate_julian;
	} elsif ($value eq "RESUME") {
	    &resume_julian;
	} elsif ($value eq "STOP") {
	    &stop_julian;
	} elsif ($value eq "SYNCGRAM") {
	    &syncgram_julian;
	} else {		# invalid value
	    print_tell("invalid value $value\n");
	    return 0;
	}
    } else {			# unknown slot
	print_tell("invalid slot $slot\n");
	return 0;
    }
}

#############################################################
#
# handle "inq" command
# 
# return value: result string, "" for error (unknown slot)
#
sub SubCall_Inq{
    my($slot) = @_;
    my($ret);
    my($str1, $str2);

    # common inquery for all modules
    if ($slot eq "Run") {
	$ret = $module_status;
    } elsif ($commoninq{$slot} ne "") {
	$ret = $commoninq{$slot};
    # SRM-specific inq
    } elsif ($slot eq "Run.process") {
	if (defined($julian_socket)) { # julian is running
	    send_command_julian("STATUS"); # tell julian to respond
	    $ret = "";
	} else {		# julian is dead
	    $ret = "SLEEP";
	}
    } elsif ($slot eq "Grammar.list") {
	send_command_julian("GRAMINFO"); # tell julian to respond
    } elsif ($slot eq "Options") { # return all julian options
	$ret = &optval2comarg;
    } elsif ($slot eq "Julian") { # return julian binary to be executed
	$ret = &binval2binname;
    } elsif ($slot eq "Julius") { # return julius binary to be executed
	$ret = &binval2binname;
    } elsif ($inq2bin{$slot} ne "") {
	$ret = $binval{$inq2bin{$slot}};
    } elsif ($inq2opt{$slot} ne "") {
	if ($set2opt{$slot} eq "??") { # not implemented yet
	    print_tell("not implemented slot $slot\n");
	    $ret = "";
	} else {
	    if ($SLOT{$slot} ne "") {
		# return slot value
		$ret = $SLOT{$slot};
	    } else {
		# return real option value
		if ($inq2opt{$slot} =~ /^(.*):sw_(.)/) {
		    $str1 = $1; $str2 = $2;
		    if ($optval{$str1} =~ /$str2/) {
			$ret = "ON";
		    } else {
			$ret = "OFF";
		    }
		} else {
		    $ret = $optval{$inq2opt{$slot}};
		    if ($ret eq "") {
			$ret = "OFF";
		    }
		}
	    }

	}
    } else {			# unknown slot
	print_tell("invalid slot $slot\n");
	$ret = "";
    }
    return($ret);
}

#############################################################
#############################################################
#
# control Julian
#
# communicate with recognition engine via socket
#
# command and response are handled asynchronously:
#   SRM_Julian.pl -> Julian: use "send_command_julian(COMMAND)"
#   Julian -> SRM_Julian.pl: handled in "SubCall_ProcessOutput()"
#            (output socket polled at process_input() in SRM_Main.pl)
# 
#############################################################
#############################################################

#############################################################
#
# process and re-form output of recognition engine
# (called from SRM_Main.pl)
#
sub SubCall_ProcessOutput {
    my($str, $allstr);
    my($sysinfo, $sysmes);
    # receive string (allow multiple commands)
    $allstr = "";
    while ( $allstr !~ /\n\.\n/) {
	while(<$julian_socket>) {
	    $allstr .= $_;
	}
    }
    foreach $str (split(/\n\.\n/, $allstr)) {
	# process the output
	if ($str =~ /<SYSINFO PROCESS=\"(.*)\"( MESSAGE=\"(.*)\")?\/>/) { # status output
	    $sysinfo = $1; $sysmes = $3;
	    if ($sysinfo eq "ERREXIT") {
		print_tell("recognizer exited by error: $sysmes\n");
		&close_julian;
	    } else {
		print_rep("$sysinfo\n");
	    }
	} else {			# recognition result and input status
	    chomp($str);
	    if ($str =~ /INPUT STATUS=\"LISTEN\"/) {
		if ($module_started == 0) {
		    $module_started = 1;
		    print_tell("ModuleStart\n");
		}
	    }
	    if ($str =~ /\n/) {
		print_tell("<< EOM\n${str}\nEOM\n");
	    } else {
		print_tell("${str}\n");
	    }
	}
    }
    
}
#############################################################
#
# return if the SRM sub-module is ready and running
# (called from SRM_Main.pl)
#
sub SubCall_OutputAlive {
    if (defined($julian_socket)) {
	return 1;
    }
    return 0;
}
#############################################################
#
# return the socket handler of SRM sub-module
# (called from SRM_Main.pl)
#
sub SubCall_OutputHandler {
    return($julian_socket);
}

#############################################################
#
# send command to Julian
#
sub send_command_julian {
    print $julian_socket "@_\n"; # commands should be separated with "\n"
}
sub send_data_julian {
    print $julian_socket "@_"; # send raw data
}
#############################################################
#
# startup julian sub-process
#
sub start_julian {
    my($comarg);
    my($oldfh);
    my($sockaddr);
    my($count);
    my($julianbin);

    if (defined($julian_socket)) { # already started
	print_tell("recognizer already running\n");
    } else {
	# select julian binary to start
	$julianbin = &binval2binname;
	# make command arguments from optval
	$comarg = &optval2comarg . "-module $jport";
	# start julian in module mode
	$comarg .= " > $logfile 2>&1" if ($debug != 1);
	$jpid = fork;
	if ($jpid == 0){
	    exec ("$julianbin $comarg");
	}
	# connect to the julian process
	$count = 0;
	while ($count < 100) {
	    $julian_socket = IO::Socket::INET->new(PeerAddr => "localhost",
						   PeerPort => $jport,
						   Proto     => 'tcp',
						   TimeOut   => '5',
						   Reuse     => 1);
	    last if ($julian_socket);
	    #sleep(0.05);
	    select(undef,undef,undef,0.05);
	    $count ++;
	}
	if (! $julian_socket) {	# time out (0.05 x 100 = 5 sec)
	    print_tell("cannot connect, engine startup failed\n");
	    return;
	}
	# set non-blocking
	$oldfh = select($julian_socket); $| = 1; select($oldfh);
	unless (fcntl($julian_socket, &F_SETFL, O_NONBLOCK)) {
	    die "can't set as nonblock\n";
	}
	$module_started = 0;
    }
}
#############################################################
#
# close connection
#
sub close_julian {
    # clear socket handler
    close($julian_socket);
    undef($julian_socket);
    $module_started = 0;
    print_tell("ModuleExit\n");
}

#############################################################
#
# grammar sync
#
sub syncgram_julian {
    send_command_julian("SYNCGRAM");
}

#############################################################
#
# kill the julian sub-process
#
sub stop_julian {
    # kill julian
    if (defined($julian_socket)) {
	# send 'die' command to julian
	send_command_julian("DIE");
	close($julian_socket);
	undef($julian_socket);
    }
    $module_started = 0;
    print_tell("ModuleExit\n");
}
#############################################################
#
# julian command: pause recognition
# if recognition is running at that time, wait till the recognition finishes
#
sub pause_julian {
    send_command_julian("PAUSE");
}
#############################################################
#
# julian command: terminate recognition
# recognition will be paused as soon as this command is received
#
sub terminate_julian {
    send_command_julian("TERMINATE");
}
#############################################################
#
# julian command: resume recognition
# resume the paused/terminated recognition process
#
sub resume_julian {
    send_command_julian("RESUME");
}
#############################################################
#
# julian command: change grammar file
#
sub change_gram_julian {
    my($optstr, $type, $gram, $dic);

    ($optstr, $type) = split(/:/, $set2opt{"Grammar"});
    $gram = $optval{$optstr};
    ($optstr, $type) = split(/:/, $set2opt{"Dic"});
    $dic = $optval{$optstr};

    if ($modified{"grammar"} == 1 && $modified{"dict"} == 1) {
	print_tell("grammar/dict is now being sent to Julian...\n");
	if (! open(GRAM, "$gram")) {
	    print_tell("grammar file \"$gram\" not found\n");
	    return 0;
	}
	if (! open(DICT, "$dic")) {
	    print_tell("dict file \"$dic\" not found\n");
	    close(GRAM);
	    return 0;
	}
	send_command_julian("CHANGEGRAM");
	while(<GRAM>) {
	    send_data_julian($_);
	}
	send_data_julian("DFAEND\n");
	while(<DICT>) {
	    send_data_julian($_);
	}
	send_data_julian("DICEND\n");
	print_tell("grammar send complete\n");
	$modified{"grammar"} = 0;
	$modified{"dict"}    = 0;
    }
}

############################################################
#
# make julian executable name from $binval{}
#
sub binval2binname {
    my ($suffix);
    my ($binopt);
    $suffix = $julian_suffix_template;
    foreach $binopt (keys %binval) {
	$suffix =~ s/$binopt/$binval{$binopt}/g;
    }
    return $julianbin_prefix . $suffix;
}
    
############################################################
#
# make option argument from $optval{}
#
sub optval2comarg {
    my ($com);
    my ($opt);
    my ($incode, $outcode);

    $com = "";
    $incode = "";
    $outcode = "";
    foreach $opt (keys %optval) {
	# 06/02/03 by ri: charconv
	if ($opt eq "-charconv_in") {
	    $incode = $optval{$opt};
	    next;
	}
	# 06/02/03 by ri: charconv
	if ($opt eq "-charconv_out") {
	    $outcode = $optval{$opt};
	    next;
	}
	if ($optval{$opt} eq "OFF") {next;}
	if ($optval{$opt} eq "ON") {
	    $com .= $opt . " ";
	} else {
	    $com .= $opt . " ". $optval{$opt} . " ";
	    # special: add same argument of "-n" to "-output"
	    if ($opt eq "-n") {
		$com .= "-output" . " ". $optval{$opt} . " ";
	    }
	}
    }
    # 06/02/03 by ri: charconv
    if ($outcode ne "" && $outcode ne "noconv") {
	# conversion explicitly specified
	# we have to know the charset of the Grammar
	if ($incode eq "" || $incode eq "default") {
	    # Grammar charset not specified
	    # assume EUC
	    $incode = "euc-jp";
	}
	if ($incode eq "euc") {
	    $incode = "euc-jp";
	}
	if ($incode ne $outcode) {
	    $com .= "-charconv $incode $outcode ";
	}
    }

    return $com;
}

1;
## end of script ###########################################
